=== modified file 'README.txt'
--- README.txt	2008-06-16 08:49:28 +0000
+++ README.txt	2008-03-03 23:10:35 +0000
@@ -56,12 +56,20 @@
 ------
 
 To speed up operations with large branches, loggerhead can be configured
-to create a files changed cache per branch on request:
+to create three caches per branch when it first launches:
 
+    - a revision data cache
     - a cache of files changed in a revision
+    - a text searching cache
 
 You can put the cache folder anywhere.
 
+The revision data cache and text searching cache will be automatically
+filled once loggerhead starts up.  The caches for a branch with 15,000
+revisions take about four minutes each on a fairly fast machine, but
+mileage may vary.  Once they are built, they update every six hours or
+so but usually finish quickly (or instantly) after the initial creation.
+
 
 SUPPORT
 -------

=== modified file 'loggerhead.conf.example'
--- loggerhead.conf.example	2008-06-16 08:49:28 +0000
+++ loggerhead.conf.example	2008-03-06 07:30:37 +0000
@@ -27,8 +27,8 @@
     # for the top browsing page, you can include a description, in raw HTML.
     #description = 'Bazaar is a <i>revision control system</i>.'
 
-    # should we cache files changed? (recommended)
-    # this will be a folder containing one file.  if the folder doesn't
+    # should we cache revisions and build a text index?  (HIGHLY recommended)
+    # this will be a folder containing a few files.  if the folder doesn't
     # exist, it'll be created.
     #
     # all branches in this project will share the same cache, which is usually

=== modified file 'loggerhead/branchview.py'
--- loggerhead/branchview.py	2008-06-16 08:49:28 +0000
+++ loggerhead/branchview.py	2008-04-18 00:33:19 +0000
@@ -29,8 +29,9 @@
 from cherrypy import HTTPRedirect
 
 from loggerhead import util
-from loggerhead.changecache import FileChangeCache
+from loggerhead.changecache import ChangeCache, FileChangeCache
 from loggerhead.history import History
+from loggerhead.textindex import TextIndex
 from loggerhead.controllers.changelog_ui import ChangeLogUI
 from loggerhead.controllers.atom_ui import AtomUI
 from loggerhead.controllers.revision_ui import RevisionUI
@@ -71,6 +72,17 @@
         # force history object to be loaded:
         self.get_history()
 
+        turbogears.startup.call_on_shutdown.append(self.close)
+
+    @with_history_lock
+    def close(self):
+        # it's important that we cleanly detach the history, so the cache
+        # files can be closed correctly and hopefully remain uncorrupted.
+        # this should also stop any ongoing indexing.
+        self._history.detach()
+        self._history = None
+        self._closed = True
+
     config = property(lambda self: self._config)
 
     name = property(lambda self: self._name)
@@ -135,6 +147,8 @@
             return None
         if (self._history is None) or self._history.out_of_date():
             self.log.debug('Reload branch history...')
+            if self._history is not None:
+                self._history.detach()
             _history = self._history = History.from_folder(
                 self._absfolder, self._name)
             cache_path = self._config.get('cachepath', None)
@@ -142,10 +156,15 @@
                 # try the project config
                 cache_path = self._project_config.get('cachepath', None)
             if cache_path is not None:
+                _history.use_cache(ChangeCache(_history, cache_path))
                 _history.use_file_cache(FileChangeCache(_history, cache_path))
+                _history.use_search_index(TextIndex(_history, cache_path))
         return self._history
 
+    def check_rebuild(self):
         h = self.get_history()
+        if h is not None:
+            h.check_rebuild()
 
     def url(self, elements, **kw):
         "build an url relative to this branch"

=== modified file 'loggerhead/changecache.py'
--- loggerhead/changecache.py	2008-06-16 09:07:39 +0000
+++ loggerhead/changecache.py	2008-02-27 03:36:09 +0000
@@ -50,6 +50,8 @@
                 ).replace('?', _param_marker)
 _insert_stmt = ("insert into revisiondata (revid, data) "
                 "values (?, ?)").replace('?', _param_marker)
+_update_stmt = ("update revisiondata set data = ? where revid = ?"
+                ).replace('?', _param_marker)
 
 
 
@@ -83,6 +85,11 @@
             self.cursor.execute(_insert_stmt, (r, self._serialize(d)))
         if commit:
             self.connection.commit()
+    def update(self, revid_obj_pairs, commit=True):
+        for  (r, d) in revid_obj_pairs:
+            self.cursor.execute(_update_stmt, (self._serialize(d), r))
+        if commit:
+            self.connection.commit()
     def count(self):
         self.cursor.execute(
             "select count(*) from revisiondata")
@@ -92,6 +99,124 @@
             self.connection.commit()
         self.connection.close()
 
+class ChangeCache (object):
+
+    def __init__(self, history, cache_path):
+        self.history = history
+        self.log = history.log
+
+        if not os.path.exists(cache_path):
+            os.mkdir(cache_path)
+
+        self._changes_filename = os.path.join(cache_path, 'changes.sql')
+
+        # use a lockfile since the cache folder could be shared across different processes.
+        self._lock = LockFile(os.path.join(cache_path, 'lock'))
+        self._closed = False
+
+##         # this is fluff; don't slow down startup time with it.
+##         # but it is racy in tests :(
+##         def log_sizes():
+##             self.log.info('Using change cache %s; %d entries.' % (cache_path, self.size()))
+##         threading.Thread(target=log_sizes).start()
+
+    def _cache(self):
+        return FakeShelf(self._changes_filename)
+
+    @with_lock
+    def close(self):
+        self.log.debug('Closing cache file.')
+        self._closed = True
+
+    @with_lock
+    def closed(self):
+        return self._closed
+
+    @with_lock
+    def flush(self):
+        pass
+
+    @with_lock
+    def get_changes(self, revid_list):
+        """
+        get a list of changes by their revision_ids.  any changes missing
+        from the cache are fetched by calling L{History.get_change_uncached}
+        and inserted into the cache before returning.
+        """
+        out = []
+        missing_revids = []
+        missing_revid_indices = []
+        cache = self._cache()
+        for revid in revid_list:
+            entry = cache.get(revid)
+            if entry is not None:
+                out.append(entry)
+            else:
+                missing_revids.append(revid)
+                missing_revid_indices.append(len(out))
+                out.append(None)
+        if missing_revids:
+            missing_entries = self.history.get_changes_uncached(missing_revids)
+            missing_entry_dict = {}
+            for entry in missing_entries:
+                missing_entry_dict[entry.revid] = entry
+            revid_entry_pairs = []
+            for i, revid in zip(missing_revid_indices, missing_revids):
+                out[i] = entry = missing_entry_dict.get(revid)
+                if entry is not None:
+                    revid_entry_pairs.append((revid, entry))
+            cache.add(revid_entry_pairs)
+        return filter(None, out)
+
+    @with_lock
+    def full(self):
+        cache = self._cache()
+        last_revid = util.to_utf8(self.history.last_revid)
+        revision_history = self.history.get_revision_history()
+        return (cache.count() >= len(revision_history)
+                and cache.get(last_revid) is not None)
+
+    @with_lock
+    def size(self):
+        return self._cache().count()
+
+    def check_rebuild(self, max_time=3600):
+        """
+        check if we need to fill in any missing pieces of the cache.  pull in
+        any missing changes, but don't work any longer than C{max_time}
+        seconds.
+        """
+        if self.closed() or self.full():
+            return
+
+        self.log.info('Building revision cache...')
+        start_time = time.time()
+        last_update = time.time()
+        count = 0
+
+        work = list(self.history.get_revision_history())
+        jump = 100
+        for i in xrange(0, len(work), jump):
+            r = work[i:i + jump]
+            # must call into history so we grab the branch lock (otherwise, lock inversion)
+            self.history.get_changes(r)
+            if self.closed():
+                self.flush()
+                return
+            count += jump
+            now = time.time()
+            if now - start_time > max_time:
+                self.log.info('Cache rebuilding will pause for now.')
+                self.flush()
+                return
+            if now - last_update > 60:
+                self.log.info('Revision cache rebuilding continues: %d/%d' % (min(count, len(work)), len(work)))
+                last_update = time.time()
+                self.flush()
+            # give someone else a chance at the lock
+            time.sleep(1)
+        self.log.info('Revision cache rebuild completed.')
+        self.flush()
 
 class FileChangeCache(object):
     def __init__(self, history, cache_path):

=== modified file 'loggerhead/controllers/__init__.py'
--- loggerhead/controllers/__init__.py	2008-06-16 08:49:28 +0000
+++ loggerhead/controllers/__init__.py	2008-05-28 22:47:44 +0000
@@ -117,6 +117,12 @@
         vals.update(templatefunctions)
         return vals
 
+    def _check_rebuild(self):
+        for p in self._projects:
+            for v in p.views:
+                v.check_rebuild()
+
+
 
 # for use in profiling the very-slow get_change() method:
 #h = Root.bazaar.bzr_dev.get_history()

=== modified file 'loggerhead/controllers/annotate_ui.py'
--- loggerhead/controllers/annotate_ui.py	2008-06-16 08:49:28 +0000
+++ loggerhead/controllers/annotate_ui.py	2008-06-11 05:26:33 +0000
@@ -93,6 +93,7 @@
                 'url': self._branch.context_url,
             }
             vals.update(templatefunctions)
+            h.flush_cache()
             self.log.info('/annotate: %r secs' % (time.time() - z,))
             return vals
         finally:

=== modified file 'loggerhead/controllers/atom_ui.py'
--- loggerhead/controllers/atom_ui.py	2008-06-16 08:49:28 +0000
+++ loggerhead/controllers/atom_ui.py	2008-06-10 04:48:26 +0000
@@ -50,6 +50,7 @@
                 'updated': entries[0].date.isoformat() + 'Z',
             }
             vals.update(templatefunctions)
+            h.flush_cache()
             return vals
         finally:
             h._branch.unlock()

=== modified file 'loggerhead/controllers/changelog_ui.py'
--- loggerhead/controllers/changelog_ui.py	2008-06-16 08:49:28 +0000
+++ loggerhead/controllers/changelog_ui.py	2008-06-11 06:07:23 +0000
@@ -57,7 +57,6 @@
             try:
                 revid, start_revid, revid_list = h.get_view(
                     revid, start_revid, filter_file_id, query)
-                kw['start_revid'] = start_revid
                 util.set_context(kw)
 
                 if (query is not None) and (len(revid_list) == 0):
@@ -102,7 +101,7 @@
                         break
 
             def url(pathargs, **kw):
-                return self._branch.url(pathargs, **util.get_context(**kw))
+                return self._branch.url(pathargs, **kw)
 
             vals = {
                 'branch': self._branch,
@@ -120,6 +119,7 @@
                 'url': self._branch.context_url,
             }
             vals.update(templatefunctions)
+            h.flush_cache()
             self.log.info('/changes %r: %r secs' % (revid, time.time() - z))
             return vals
         finally:

=== modified file 'loggerhead/controllers/inventory_ui.py'
--- loggerhead/controllers/inventory_ui.py	2008-06-16 08:49:28 +0000
+++ loggerhead/controllers/inventory_ui.py	2008-06-11 05:26:33 +0000
@@ -105,6 +105,7 @@
                 'start_revid': start_revid,
             }
             vals.update(templatefunctions)
+            h.flush_cache()
             self.log.info('/inventory %r: %r secs' % (revid, time.time() - z))
             return vals
         finally:

=== modified file 'loggerhead/controllers/revision_ui.py'
--- loggerhead/controllers/revision_ui.py	2008-06-16 08:49:28 +0000
+++ loggerhead/controllers/revision_ui.py	2008-06-11 05:26:33 +0000
@@ -106,6 +106,7 @@
                 'show_plain_diffs': line_count > line_count_limit,
             }
             vals.update(templatefunctions)
+            h.flush_cache()
             self.log.info('/revision: %r seconds' % (time.time() - z,))
             return vals
         finally:

=== modified file 'loggerhead/history.py'
--- loggerhead/history.py	2008-06-16 08:49:28 +0000
+++ loggerhead/history.py	2008-06-11 04:41:08 +0000
@@ -191,7 +191,9 @@
 class History (object):
 
     def __init__(self):
+        self._change_cache = None
         self._file_change_cache = None
+        self._index = None
         self._lock = threading.RLock()
 
     @classmethod
@@ -275,13 +277,44 @@
             return True
         return self._branch.last_revision() != self._last_revid
 
+    def use_cache(self, cache):
+        self._change_cache = cache
+
     def use_file_cache(self, cache):
         self._file_change_cache = cache
 
+    def use_search_index(self, index):
+        self._index = index
+
     @property
     def has_revisions(self):
         return not bzrlib.revision.is_null(self.last_revid)
 
+    @with_branch_lock
+    def detach(self):
+        # called when a new history object needs to be created, because the
+        # branch history has changed.  we need to immediately close and stop
+        # using our caches, because a new history object will be created to
+        # replace us, using the same cache files.
+        # (may also be called during server shutdown.)
+        if self._change_cache is not None:
+            self._change_cache.close()
+            self._change_cache = None
+        if self._index is not None:
+            self._index.close()
+            self._index = None
+
+    def flush_cache(self):
+        if self._change_cache is None:
+            return
+        self._change_cache.flush()
+
+    def check_rebuild(self):
+        if self._change_cache is not None:
+            self._change_cache.check_rebuild()
+        #if self._index is not None:
+        #    self._index.check_rebuild()
+
     last_revid = property(lambda self: self._last_revid, None, None)
 
     @with_branch_lock
@@ -365,6 +398,17 @@
         self.log.debug('searched %d revisions for %r in %r secs', len(revid_list), text, time.time() - z)
         return out
 
+    def get_revision_history_matching_indexed(self, revid_list, text):
+        self.log.debug('searching %d revisions for %r', len(revid_list), text)
+        z = time.time()
+        if self._index is None:
+            return self.get_revision_history_matching(revid_list, text)
+        out = self._index.find(text, revid_list)
+        self.log.debug('searched %d revisions for %r in %r secs: %d results', len(revid_list), text, time.time() - z, len(out))
+        # put them in some coherent order :)
+        out = [r for r in self._full_history if r in out]
+        return out
+
     @with_branch_lock
     def get_search_revid_list(self, query, revid_list):
         """
@@ -407,6 +451,11 @@
                 revid_list = list(self.get_revids_from(None, self._last_revid))
             return self.get_revision_history_since(revid_list, date)
 
+        # check comment fields.
+        if revid_list is None:
+            revid_list = self._full_history
+        return self.get_revision_history_matching_indexed(revid_list, query)
+
     revno_re = re.compile(r'^[\d\.]+$')
     # the date regex are without a final '$' so that queries like
     # "2006-11-30 12:15" still mostly work.  (i think it's better to give
@@ -490,13 +539,12 @@
         else:
             revid_list = None
 
-        try:
-            revid_list = self.get_search_revid_list(query, revid_list)
-            if len(revid_list) > 0:
-                if revid not in revid_list:
-                    revid = revid_list[0]
-                return revid, start_revid, revid_list
-        except:
+        revid_list = self.get_search_revid_list(query, revid_list)
+        if len(revid_list) > 0:
+            if revid not in revid_list:
+                revid = revid_list[0]
+            return revid, start_revid, revid_list
+        else:
             # no results
             return None, None, []
 
@@ -599,7 +647,10 @@
 
         Revisions not present and NULL_REVISION will be ignored.
         """
-        changes = self.get_changes_uncached(revid_list)
+        if self._change_cache is None:
+            changes = self.get_changes_uncached(revid_list)
+        else:
+            changes = self._change_cache.get_changes(revid_list)
         if len(changes) == 0:
             return changes
 
@@ -609,8 +660,11 @@
             merge_revids = self.simplify_merge_point_list(self.get_merge_point_list(change.revid))
             change.merge_points = [util.Container(revid=r, revno=self.get_revno(r)) for r in merge_revids]
             if len(change.parents) > 0:
-                change.parents = [util.Container(revid=r, 
-                    revno=self.get_revno(r)) for r in change.parents]
+                if isinstance(change.parents[0], util.Container):
+                    # old cache stored a potentially-bogus revno
+                    change.parents = [util.Container(revid=p.revid, revno=self.get_revno(p.revid)) for p in change.parents]
+                else:
+                    change.parents = [util.Container(revid=r, revno=self.get_revno(r)) for r in change.parents]
             change.revno = self.get_revno(change.revid)
 
         parity = 0
@@ -634,14 +688,16 @@
     @with_branch_lock
     @with_bzrlib_read_lock
     def get_changes_uncached(self, revid_list):
+        # FIXME: deprecated method in getting a null revision
         revid_list = filter(lambda revid: not bzrlib.revision.is_null(revid),
                             revid_list)
-        parent_map = self._branch.repository.get_graph().get_parent_map(revid_list)
+        repo = self._branch.repository
+        parent_map = repo.get_graph().get_parent_map(revid_list)
         # We need to return the answer in the same order as the input,
         # less any ghosts.
         present_revids = [revid for revid in revid_list
                           if revid in parent_map]
-        rev_list = self._branch.repository.get_revisions(present_revids)
+        rev_list = repo.get_revisions(present_revids)
 
         return [self._change_from_revision(rev) for rev in rev_list]
 

=== modified file 'loggerhead/templatefunctions.py'
--- loggerhead/templatefunctions.py	2008-06-11 00:02:50 +0000
+++ loggerhead/templatefunctions.py	2008-06-11 04:41:08 +0000
@@ -22,11 +22,11 @@
         **templatefunctions)
 
 @templatefunc
-def revisioninfo(url, branch, entry, modified_file_link=None):
+def revisioninfo(url, branch, entry, revid, modified_file_link=None):
     from loggerhead import util
     return _pt('revisioninfo').expand(
         url=url, change=entry, branch=branch, util=util,
-        modified_file_link=modified_file_link,
+        modified_file_link=modified_file_link, revid=revid,
         **templatefunctions)
 
 @templatefunc

=== modified file 'loggerhead/templates/annotate.pt'
--- loggerhead/templates/annotate.pt	2008-05-29 04:43:39 +0000
+++ loggerhead/templates/annotate.pt	2008-06-11 05:11:05 +0000
@@ -26,19 +26,19 @@
       <div class="links">
         <div>
           <b>&#8594;</b>
-          <a tal:attributes="href python:url(['/files', revid], clear=1)">
+          <a tal:attributes="href python:url(['/files', revid])">
             browse files
           </a>
         </div>
         <div>
           <b>&#8594;</b>
-          <a tal:attributes="href python:url(['/revision'], clear=1, start_revid=revid)">
+          <a tal:attributes="href python:url(['/revision'], start_revid=revid)">
             view revision
           </a>
         </div>
         <div>
           <b>&#8594;</b>
-          <a tal:attributes="href python:url(['/changes'], clear=1, start_revid=revid, filter_file_id=file_id)">
+          <a tal:attributes="href python:url(['/changes'], start_revid=revid, filter_file_id=file_id)">
             view changes to this file
           </a>
         </div>
@@ -69,7 +69,7 @@
             <td tal:attributes="class string:revno ${line/status}">
               <a tal:condition="python:line.status=='changed'"
                  tal:content="python:util.trunc(line.change.revno)"
-                 tal:attributes="href python:url(['/revision'], clear=1, start_revid=line.change.revid, filter_file_id=file_id);
+                 tal:attributes="href python:url(['/revision'], start_revid=line.change.revid, filter_file_id=file_id);
                                  title python:'%s by %s, on %s (%s)'%(line.change.revno, util.hide_email(line.change.author), line.change.date.strftime('%d %b %Y %H:%M'), util.date_time(line.change.date))">
 
                 ${util.trunc(line.change.revno)}

=== modified file 'loggerhead/templates/atom.pt'
--- loggerhead/templates/atom.pt	2008-06-10 04:48:26 +0000
+++ loggerhead/templates/atom.pt	2008-06-11 04:41:08 +0000
@@ -14,7 +14,7 @@
     <updated tal:content="entry/date/isoformat">
     Z
     </updated>
-    <id tal:content="python:branch.url(['revision', entry.revid])">
+    <id tal:content="python:branch.url(['revision', entry.revno])">
       ID
     </id>
     <author>
@@ -26,6 +26,6 @@
       comment
     </content>
     <link rel="alternate"
-          tal:attributes="href python:branch.url(['revision', entry.revid])" />
+          tal:attributes="href python:branch.url(['revision', entry.revno])" />
   </entry>
 </feed>

=== modified file 'loggerhead/templates/changelog.pt'
--- loggerhead/templates/changelog.pt	2008-06-11 04:51:53 +0000
+++ loggerhead/templates/changelog.pt	2008-06-11 05:26:33 +0000
@@ -78,7 +78,7 @@
             <tr tal:attributes="class string:revision-header parity${entry/parity}">
               <td class="revision-number">
                 <a tal:attributes="title python:'Show revision '+history.get_revno(entry.revid);
-                                   href  python:url(['/revision', entry.revid])"
+                                   href  python:url(['/revision', history.get_revno(entry.revid)])"
                    tal:content="python:util.trunc(entry.revno)">
                   text
                 </a>
@@ -88,7 +88,7 @@
               </td>
               <td class="summary">
                 <a tal:attributes="title python:'Show revision '+history.get_revno(entry.revid);
-                                   href  python:url(['/revision', entry.revid])"
+                                   href  python:url(['/revision', history.get_revno(entry.revid)])"
                    tal:content="entry/short_comment">
                   text
                 </a>
@@ -101,7 +101,7 @@
                         tal:content="python:util._approximatedate(entry.date)"></span>
               </td>
               <td class="inventory-link">
-                <a tal:attributes="href python:branch.url(['/files', entry.revid]);
+                <a tal:attributes="href python:branch.url(['/files', entry.revno]);
                                    title string:Files at revision ${entry/revno}">
                   files
                 </a>
@@ -121,7 +121,7 @@
                     </th>
                     <td class="children">
                       <a title="Show history"
-                         tal:attributes="href python:url(['/changes'], start_revid=merge_point.revid)"
+                         tal:attributes="href python:url(['/changes'], start_revid=merge_point.revno)"
                          tal:content="python:revno_with_nick(merge_point)">
                         class="revlink">
                         ${text}
@@ -136,7 +136,7 @@
                     </th>
                     <td class="parents">
                       <a title="Show history"
-                         tal:attributes="href python:url(['/changes'], start_revid=parent.revid)"
+                         tal:attributes="href python:url(['/changes'], start_revid=parent.revno)"
                          tal:content="python:revno_with_nick(parent)">
                         class="revlink">
                         ${text}

=== modified file 'loggerhead/templates/inventory.pt'
--- loggerhead/templates/inventory.pt	2008-06-11 04:53:47 +0000
+++ loggerhead/templates/inventory.pt	2008-06-11 05:26:33 +0000
@@ -20,15 +20,15 @@
 
     <div metal:fill-slot="content">
       <div class="links">
-        <div> <b>&#8594;</b> <a tal:attributes="href python:url(['/revision'], clear=1, start_revid=revid)">
+        <div> <b>&#8594;</b> <a tal:attributes="href python:url(['/revision', change.revno])">
 	    view revision </a>
         </div>
-        <div> <b>&#8594;</b> <a tal:attributes="href python:url(['/changes'], clear=1, start_revid=revid)">
+        <div> <b>&#8594;</b> <a tal:attributes="href python:url(['/changes', change.revno])">
 	    view branch changes </a>
         </div>
       </div>
 
-      <tal:revision-info replace="structure python:revisioninfo(url, branch, change)" />
+      <tal:revision-info replace="structure python:revisioninfo(url, branch, change, revid)" />
 
       <div class="inventory-path">
         <b>folder:</b> <span class="folder" tal:content="path"> ${path} </span>
@@ -81,7 +81,7 @@
           </td>
           <td tal:attributes="class string:filename ${file/kind}">
             <a tal:condition="python:file.kind=='directory'"
-               tal:attributes="href python:url(['/files', revid], file_id=file.file_id)"
+               tal:attributes="href python:url(['/files', file.change.revno], file_id=file.file_id)"
                tal:content="string:${file/filename}/">
               ${file.filename}/
             </a>
@@ -90,14 +90,14 @@
               ${file.filename}@
             </span>
             <a tal:condition="python:file.kind=='file'"
-               tal:attributes="href python:url(['/annotate', revid], file_id=file.file_id);
+               tal:attributes="href python:url(['/annotate', file.change.revno], file_id=file.file_id);
                                title string:Annotate ${file/filename}"
                tal:content="string:${file/filename}">
               ${file.filename}
             </a>
           </td>
           <td class="revision">
-            <a tal:attributes="href python:url(['/revision', file.revid], start_revid=file.revid, file_id=file.file_id);
+            <a tal:attributes="href python:url(['/revision', file.change.revno], filter_file_id=file.file_id);
                                title string:Show revision ${file/change/revno}"
                tal:content="python:util.trunc(file.change.revno, 15)">
               ${text}

=== modified file 'loggerhead/templates/macros.pt'
--- loggerhead/templates/macros.pt	2008-06-10 23:21:43 +0000
+++ loggerhead/templates/macros.pt	2008-06-11 05:11:05 +0000
@@ -22,8 +22,8 @@
           <tr>
             <td>
               <span class="buttons">
-                <a tal:attributes="href python:url('/changes', clear=1)"> changes </a>
-                <a tal:attributes="href python:url('/files', clear=1)"> files </a>
+                <a tal:attributes="href python:url('/changes')"> changes </a>
+                <a tal:attributes="href python:url('/files')"> files </a>
                 <span class="search">
                   search: <input type="text" name="q" />
                 </span>

=== modified file 'loggerhead/templates/modified-file-link-log.pt'
--- loggerhead/templates/modified-file-link-log.pt	2008-06-11 02:19:53 +0000
+++ loggerhead/templates/modified-file-link-log.pt	2008-06-11 05:11:05 +0000
@@ -1,5 +1,5 @@
 <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
-<a tal:attributes="href python:'javascript:diff_url(\'' + url(['/revision', entry.revid]) + '#' + item.filename + '\')';
+<a tal:attributes="href python:'javascript:diff_url(\'' + url(['/revision', entry.revno]) + '#' + item.filename + '\')';
                    title string:Annotate ${item/filename}"
    tal:content="item/filename" class="filename">
   item/filename

=== modified file 'loggerhead/templates/revision.pt'
--- loggerhead/templates/revision.pt	2008-06-11 03:17:41 +0000
+++ loggerhead/templates/revision.pt	2008-06-11 04:41:08 +0000
@@ -77,7 +77,7 @@
           </tal:we-are-comparing>
         </div>
 
-        <tal:revision-info replace="structure python:revisioninfo(url, branch, change, modified_file_link_rev)" />
+        <tal:revision-info replace="structure python:revisioninfo(url, branch, change, revid, modified_file_link_rev)" />
 
         <table class="diff-option-buttons">
           <tr>

=== modified file 'loggerhead/templates/revisioninfo.pt'
--- loggerhead/templates/revisioninfo.pt	2008-05-29 05:58:01 +0000
+++ loggerhead/templates/revisioninfo.pt	2008-06-11 04:41:08 +0000
@@ -28,8 +28,7 @@
       <td class="children">
         <a title="Show history"
            tal:attributes="title string:Show revision ${merge_point/revno};
-                           href python:url(['/revision', merge_point.revid],
-                                           clear=1, start_revid=merge_point.revid)"
+                                        href python:url(['/revision', merge_point.revno])"
            tal:content="python:revno_with_nick(merge_point)">
           ${text}
         </a>
@@ -45,7 +44,7 @@
       <td class="parents">
         <a title="Show history"
            tal:attributes="title string:Show revision ${parent/revno};
-                             href python:url(['/revision', parent.revid], clear=1, start_revid=parent.revid)"
+                           href python:url(['/revision', parent.revno])"
            tal:content="python:revno_with_nick(parent)">
           class="revlink">
           ${text}
@@ -65,6 +64,17 @@
       </td>
     </tr>
 
+    <tr>
+      <th class="revid">
+        revision id:
+      </th>
+      <td class="revid">
+          <span tal:content="revid">
+              revid
+        </span>
+      </td>
+    </tr>
+
     <tal:block condition="nocall:modified_file_link">
       <tr class="divider">
         <th>

=== added file 'loggerhead/textindex.py'
--- loggerhead/textindex.py	1970-01-01 00:00:00 +0000
+++ loggerhead/textindex.py	2008-02-27 03:36:09 +0000
@@ -0,0 +1,240 @@
+#
+# Copyright (C) 2006  Robey Pointer <robey@lag.net>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+#
+
+"""
+indexing of the comment text of revisions, for fast searching.
+
+two separate database files are created:
+
+    - recorded: revid -> 1 (if the revid is indexed)
+    - index: 3-letter substring -> list(revids)
+"""
+
+import os
+import re
+import time
+
+from loggerhead import util
+from loggerhead.lockfile import LockFile
+from loggerhead.changecache import FakeShelf
+
+# if any substring index reaches this many revids, replace the entry with
+# an ALL marker -- it's not worth an explicit index.
+ALL_THRESHOLD = 1000
+ALL = 'ALL'
+
+
+with_lock = util.with_lock('_lock')
+
+
+def normalize_string(s):
+    """
+    remove any punctuation and normalize all whitespace to a single space.
+    """
+    s = util.to_utf8(s).lower()
+    # remove apostrophes completely.
+    s = re.sub(r"'", '', s)
+    # convert other garbage into space
+    s = re.sub(r'[^\w\d]', ' ', s)
+    # compress multiple spaces into one.
+    s = re.sub(r'\s{2,}', ' ', s)
+    # and finally remove leading/trailing whitespace
+    s = s.strip()
+    return s
+
+
+class TextIndex (object):
+    def __init__(self, history, cache_path):
+        self.history = history
+        self.log = history.log
+
+        if not os.path.exists(cache_path):
+            os.mkdir(cache_path)
+
+        self._recorded_filename = os.path.join(cache_path, 'textindex-recorded.sql')
+        self._index_filename = os.path.join(cache_path, 'textindex.sql')
+
+        # use a lockfile since the cache folder could be shared across different processes.
+        self._lock = LockFile(os.path.join(cache_path, 'index-lock'))
+        self._closed = False
+
+        self.log.info('Using search index; %d entries.', len(self))
+
+    def _index(self):
+        return FakeShelf(self._index_filename)
+
+    def _recorded(self):
+        return FakeShelf(self._recorded_filename)
+
+    def _is_indexed(self, revid, recorded):
+        return recorded.get(util.to_utf8(revid)) is not None
+
+    @with_lock
+    def is_indexed(self, revid):
+        recorded = self._recorded()
+        try:
+            return self._is_indexed(revid, recorded)
+        finally:
+            recorded.close()
+
+    @with_lock
+    def __len__(self):
+        recorded = self._recorded()
+        try:
+            return recorded.count()
+        finally:
+            recorded.close()
+
+    @with_lock
+    def close(self):
+        self._closed = True
+
+    @with_lock
+    def closed(self):
+        return self._closed
+
+    @with_lock
+    def flush(self):
+        pass
+
+    @with_lock
+    def full(self):
+        recorded = self._recorded()
+        last_revid = util.to_utf8(self.history.last_revid)
+        try:
+            return (recorded.count() >= len(self.history.get_revision_history())
+                    and recorded.get(last_revid) is not None)
+        finally:
+            recorded.close()
+
+    def _index_change(self, change, recorded, index):
+        """
+        currently, only indexes the 'comment' field.
+        """
+        comment = normalize_string(change.comment)
+        if len(comment) < 3:
+            return
+        for i in xrange(len(comment) - 2):
+            sub = comment[i:i + 3]
+            orig = revid_set = index.get(sub)
+            if revid_set is None:
+                revid_set = set()
+            elif revid_set == ALL:
+                # this entry got too big
+                continue
+            revid_set.add(change.revid)
+            if len(revid_set) > ALL_THRESHOLD:
+                revid_set = ALL
+            if orig is not None:
+                index.update([(sub, revid_set)], commit=False)
+            else:
+                index.add([(sub, revid_set)], commit=False)
+
+        recorded.add([(util.to_utf8(change.revid), True)], commit=False)
+
+    @with_lock
+    def index_changes(self, revid_list):
+        recorded = self._recorded()
+        index = self._index()
+        try:
+            revid_list = [r for r in revid_list if not self._is_indexed(r, recorded)]
+            change_list = self.history.get_changes(revid_list)
+            for change in change_list:
+                self._index_change(change, recorded, index)
+        finally:
+            index.close(commit=True)
+            recorded.close(commit=True)
+
+    @with_lock
+    def find(self, text, revid_list=None):
+        index = self._index()
+        try:
+            text = normalize_string(text)
+            if len(text) < 3:
+                return []
+
+            total_set = None
+            if revid_list is not None:
+                total_set = set(revid_list)
+            seen_all = False
+
+            for i in xrange(len(text) - 2):
+                sub = text[i:i + 3]
+                revid_set = index.get(sub)
+                if revid_set is None:
+                    # zero matches, stop here.
+                    return []
+                if revid_set == ALL:
+                    # skip
+                    seen_all = True
+                    continue
+                if total_set is None:
+                    total_set = revid_set
+                else:
+                    total_set.intersection_update(revid_set)
+                if len(total_set) == 0:
+                    return []
+        finally:
+            index.close()
+
+        # tricky: if seen_all is True, one of the substring indices was ALL
+        # (in other words, unindexed), so our results are actually a superset
+        # of the exact answer.
+        #
+        # if we cared, we could do a direct match on the result set and cull
+        # out any that aren't actually matches.  for now, i'm gonna say that
+        # we DON'T care, and if one of the substrings hit ALL, there's a small
+        # chance that we'll give a few false positives.
+        return total_set
+
+    def check_rebuild(self, max_time=3600):
+        """
+        check if there are any un-indexed revisions, and if so, index them.
+        but don't spend longer than C{max_time} on it.
+        """
+        if self.closed() or self.full():
+            # all done
+            return
+
+        self.log.info('Building search index...')
+        work = list(self.history.get_revision_history())
+        start_time = time.time()
+        last_update = time.time()
+        count = 0
+
+        jump = 100
+        for i in xrange(0, len(work), jump):
+            r = work[i:i + jump]
+            self.index_changes(r)
+            if self.closed():
+                return
+
+            count += jump
+            now = time.time()
+            if now - start_time > 3600:
+                # there's no point working for hours.  eventually we might even
+                # hit the next re-index interval, which would suck mightily.
+                self.log.info('Search indexing has worked for an hour; giving up for now.')
+                return
+            if now - last_update > 60:
+                self.log.info('Search indexing continues: %d/%d' % (min(count, len(work)), len(work)))
+                last_update = time.time()
+            # give someone else a chance at the lock
+            time.sleep(1)
+        self.log.info('Search index completed.')
+        self.flush()

=== modified file 'loggerhead/util.py'
--- loggerhead/util.py	2008-06-10 23:00:35 +0000
+++ loggerhead/util.py	2008-06-11 04:41:08 +0000
@@ -469,6 +469,9 @@
 
 def get_context(**overrides):
     """
+    Soon to be deprecated.
+
+
     return a context map that may be overriden by specific values passed in,
     but only contains keys from the list of valid context keys.
 

=== modified file 'start-loggerhead.py'
--- start-loggerhead.py	2008-06-16 08:49:28 +0000
+++ start-loggerhead.py	2008-06-11 04:44:45 +0000
@@ -110,6 +110,10 @@
     
     Root = Root(config)
     
+    # re-index every 6 hours
+    index_freq = config.get('cache_rebuild_frequency', 6 * 3600)
+    turbogears.scheduler.add_interval_task(initialdelay=1, interval=index_freq, action=Root._check_rebuild)
+
     try:
         turbogears.start_server(Root)
     finally:

